#!/usr/bin/perl -w
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use strict;

use Fcntl ':mode';
use List::Util qw[min max];
use DBI;
use DBD::mysql;

$ENV{LANG} = "C";

# Global variables
our $prefix;
our $cmd_prefix;

require "$prefix/libexec/modules/utils.pl";

sub ls {
 # dir
 # mysql_host
 # mysql_login
 # mysql_pass
 # mysql_db
 # show dotted files ( 1 to true)
  my ($dir,$s,$mh,$ml,$mw,$md,$a) = @_;

  my @res;
  if ( defined $s) { 
    my $search_str = "$dir";
    $search_str .= " -name "."\'$s->{expr}\'" if ($s->{expr} ne "");
    $search_str .= " -mtime ".$s->{mtime} if (defined $s->{mtime} && $s->{mtime} ne "");
    $search_str .= " -used ".$s->{atime} if (defined $s->{atime} && $s->{atime} ne "");
 
    @res = ex("find $search_str -exec ls -Q -o -d --time-style=\"+\%s\" {} \\; 2>/dev/null");
  } else {
    my $all_str = "";
    $all_str = "-a" if $a;
    @res = ex("ls $all_str -Q -o -L --time-style=\"+\%s\" \"$dir\"");
  }
  my @list;
  foreach my $l (@res) {
    my %f; my $p;
    ($p,$f{'o'},$f{'s'},$f{'t'},$f{'f'}) = $l =~ m/([sdrwx-]+)\.? +\d+ +([^ ]+) +(\d+) +(\d+) +"(.+)"/;
    next if not defined $f{'f'};
    my ($d,$ur,$uw,$ux,$gr,$gw,$gx,$or,$ow,$ox) = $p =~ m/(.)(.)(.)(.)(.)(.)(.)(.)(.)(.)/;
    $f{'f'} =~ s/^$dir\/?//;
    $f{'s'} = int $f{'s'};
    $f{'t'} = int $f{'t'};
    $f{'d'} = 0;
    $f{'d'} = 1 if ( $d eq "d");
    $f{'r'} = 0;
    $f{'r'} += 400 if ($ur eq "r");
    $f{'r'} += 200 if ($uw eq "w");
    $f{'r'} += 100 if ($ux eq "x");
    $f{'r'} += 040 if ($gr eq "r");
    $f{'r'} += 020 if ($gw eq "w");
    $f{'r'} += 010 if ($gx eq "x");
    $f{'r'} += 004 if ($or eq "r");
    $f{'r'} += 002 if ($ow eq "w");
    $f{'r'} += 001 if ($ox eq "x");
    push @list, \%f if ($f{'f'} ne "");
  }
  if (defined $mh) {
    # Connect to mysql
    my $dsn = "DBI:mysql:database=$md;host=$mh;port=3306";
    my $dbh = DBI->connect($dsn, $ml, $mw,{'RaiseError' => 1});

    my $sth = $dbh->prepare("SELECT task_id g, unix_timestamp(dt) t, middleware m,
         concat(task_name, \" @ \", hostname) f FROM grid_tasks WHERE work_dir = \"$dir\" ORDER BY task_id");
    $sth->execute();
    while (my $ref = $sth->fetchrow_hashref()) {
      $ref->{d} = 1;
      push @list, $ref;
    }
    $sth->finish();
  }
  return \@list;
}


sub to_canonical_fname {
	my $f = $_[0] || die "Argument missing: directory name\n";
	$f =~ s/\.\.//g;
	$f =~ s/\.\///g;
	$f =~ s/\/+/\//g;
	$f =~ s/\/$//g;
	$f =~ s/\/\.$//g;
	return ($f);
}


sub chmodf {
  my ($files_ref,$mode,$rec) = @_;
  my $res = "";
  my $rec_s = "";
  $rec_s = "-R" if ($rec == 1);
  foreach my $file_data (@$files_ref) {
    my $file = to_canonical_fname($file_data);
    $res .= ex("chmod $rec_s $mode \"$file\" 2>&1");
  }
  return $res;
}


sub less {
  my ($f,$o,$s) = @_;
  my %less;
  open (my $fh, "< :encoding(UTF-8)", $f);
  if ($o >= 0) {
    my $t;
    read $fh, $t, -s $f;
    $less{filesize} = length $t;
    $less{text} = substr ($t, $o, $s);
  } else {
    seek ($fh, -1*$s, 2);
    read $fh, $less{text}, $s;
  }
#  $less{text} =~ s/"/&quot;/g;
#  $less{text} =~ s/'/&#039;/g;
#  $less{text} =~ s/</&lt;/g;
#  $less{text} =~ s/>/&gt;/g;
  return \%less;
}

1;
