#!/usr/bin/perl -w
#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use strict;

use DBI;

use Fcntl ':mode';
use List::Util qw[min max];


# Parsing Config Module
# Use: http://search.cpan.org/~tlinden/Config-General/
# Debian: libconfig-general-perl
# RHEL: perl-Config-General
use Config::General;

use Cwd 'abs_path';
use File::Basename;

$ENV{LANG} = "C";

# Global variable 
our $prefix;

# Begin parsing config
my $configfile = new Config::General("$prefix/etc/grid.conf");
my %config = $configfile->getall;

use Date::Parse;
# RHEL: perl-TimeDate.noarch

# Set Globus shared libs
# if ( defined $config{GLOBUS_LIBS} && $config{GLOBUS_LIBS} ne "") {
#  $ENV{LD_LIBRARY_PATH} = $config{GLOBUS_LIBS};
#  }

if ((defined $config{GRID_PREFIX}) && $config{GRID_PREFIX} ne "") {
  $config{GRID_PREFIX} .= "/";
} else {
  $config{GRID_PREFIX} = "";
}

if ((defined $config{GLOBUS_PREFIX}) && $config{GLOBUS_PREFIX} ne "") {
  $config{GLOBUS_PREFIX} .= "/";
} else {
  $config{GLOBUS_PREFIX} = "";
}



# --------------------------- Functions ---------------------------------------------------

sub grid_cert_install {
  # cert - certificate
  my ($cert,$key) = @_;
  system ("mkdir -p $ENV{HOME}/.globus");
  if ((defined $cert) && $cert ne "") {
    open (CERT, ">", "$ENV{HOME}/.globus/usercert.pem");
    print CERT $cert;
    close CERT;
  }
  if ((defined $key) && $key ne "") {
    open (KEY, ">", "$ENV{HOME}/.globus/userkey.pem");
    print KEY $key;
    close KEY;
  }
  system ("chmod -R 700 $ENV{HOME}/.globus");
  system ("chmod -R 600 $ENV{HOME}/.globus/*");
  return grid_cert_diag();
}


sub grid_vo_install {
  # Install virtual organization list
  my ($conf) = @_;
  system ("mkdir -p $ENV{HOME}/.voms");
  open (VO, ">", "$ENV{HOME}/.voms/voms.conf");
  print VO $conf;
  close VO;
  return grid_vo_list();
}

sub grid_vo_list {
  # List names and contence of virtual organizations files
  my @vo_l;
  foreach my $l (`[ -f $ENV{HOME}/.voms/voms.conf ] && cat $ENV{HOME}/.voms/voms.conf`) {
    my %vo_h;
    ($vo_h{alias},$vo_h{server},$vo_h{port},$vo_h{dn},$vo_h{name}) =
      $l =~ m/"(.+)" +"(.+)" +"(.+)" +"(.+)" +"(.+)"/;
    push @vo_l, \%vo_h if defined $vo_h{alias};
  }
  return \@vo_l;
}


#echo 123456 | grid-proxy-init -pwstdin
# voms-proxy-init --voms $vo

#belous@ds4 bin]$ ll mypro*
#-rwxr-xr-x 1 root root 12448 Янв 20  2012 myproxy-change-pass-phrase
#-rwxr-xr-x 1 root root 12136 Янв 20  2012 myproxy-destroy
#-rwxr-xr-x 1 root root 18088 Янв 20  2012 myproxy-get-delegation
#-rwxr-xr-x 1 root root 12080 Янв 20  2012 myproxy-get-trustroots
#-rwxr-xr-x 1 root root 12128 Янв 20  2012 myproxy-info
#-rwxr-xr-x 1 root root 19048 Янв 20  2012 myproxy-init
#-rwxr-xr-x 1 root root 18080 Янв 20  2012 myproxy-logon
#-rwxr-xr-x 1 root root 21824 Янв 20  2012 myproxy-retrieve
#-rwxr-xr-x 1 root root 17280 Янв 20  2012 myproxy-store

sub grid_proxy_init {
  # passwd - certificate password
  my ($passwd) = @_;
  my $res ="";
  open (PROXY, "| $config{GLOBUS_PREFIX}grid-proxy-init -pwstdin > /dev/null");
  print PROXY "$passwd\n";
  close PROXY;
  return grid_proxy_info();
}


sub grid_proxy_info {
  my $res;
  my %proxy;
  require "$prefix/libexec/modules/utils.pl";
  my @res = `$config{GLOBUS_PREFIX}grid-proxy-info 2>&1`;
  foreach my $l (@res) {
    if (my ($var,$val) = $l =~ m/^ *([^: ]+) *: (.*)/) {
      if ( $var eq "timeleft") {
        $proxy{$var} = str2sec($val);
      } else {
        $proxy{$var} = $val;
      }
    }
  }
  return \%proxy;
}


sub grid_proxy_destroy {
  my $res;
  $res = `$config{GLOBUS_PREFIX}grid-proxy-destroy`;
  return $res;
}


# check certificate
sub grid_cert_diag {
  my $r = `ls ~/.globus/usercert.pem 2>/dev/null`;
  if ( $r =~ m/usercert.pem/) {
    my @r = `cat ~/.globus/usercert.pem`;
    my $res = "";
    my $stop = 0;
    my %cert;
    foreach my $l (@r) {
      $stop = 1 if ($l =~ m/--BEGIN CERTIFICATE--/);
      last if $stop;
      if (my ($p) = $l =~ m/Issuer: (.+)/) {
        $cert{'issuer'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Subject: (.+)/) {
        $cert{'person'} = $p;
        next;
      }
      if (my ($p) = $l =~ m/Not Before: (.+)/) {
        $cert{'valid_from'} = str2time($p);
        next;
      }
      if (my ($p) = $l =~ m/Not After : (.+)/) {
        $cert{'valid_to'} = str2time($p);
        next;
      }
      $cert{'key'} = 1 if -s "$ENV{HOME}/.globus/userkey.pem";
    }
    return \%cert;
  } else {
    return 0;
  }
  return 0;
}

# ngsub -e '&(executable="/bin/echo")(arguments="Hello World")(stdout="hello.txt")'
sub grid_launch {
  # Launch grif job
  # Parameters: 
  # H - cluster  
  # J - jobname
  # d - current dir
  # iF - input files
  # oF - output files   
  # a - app name                  
  # p - partition                 
  # n - number of CPUs            
  # t - time limit                        
  # r - runtime (rsl flag)
  # g - commandline args
  # mysql_host
  # mysql_login
  # mysql_pass
  # mysql_db
                                
  my ($H,$J,$d,$iF,$oF,$a,$e,$p,$n,$t,$r,$g,$mh,$ml,$mw,$md) = @_;
  my $cluster="";
  $cluster = "-c $H" if (defined $H && $H ne "");
  my $queue="";
  $queue = "(queue=$p)" if (defined $p && $p ne "");
  my $env="";
  $env = "(runtime=$e)" if (defined $e && $e ne "");
# Parse aruments
  $g =~ s/ /\" \"/g;
  $g = "\"".$g."\"";

  my $inputFiles = "";
  if (defined $iF && $iF ne ""){
    $inputFiles = "(inputFiles=";
    foreach my $i (split (",",$iF)) {
      $inputFiles .= "(\"$i\" \"\")";
    }
    $inputFiles .= ")";
  }

  my $outputFiles = "";
  if (defined $oF && $oF ne ""){
    $outputFiles = "(outputFiles=";
    foreach my $i (split (",",$oF)) {
      $outputFiles .= "(\"$i\" \"\")";
    }
    $outputFiles .= ")";
  }
	
  chdir ($d);
  my %res;
  my $res_str;
  if (defined $r && $r eq "rsl") {
# $r = `nohup $config{GRID_PREFIX}/ngsub $cluster -f \'$a\' 2>&1 /dev/null &`;
    $res_str = `$config{GRID_PREFIX}ngsub $cluster -f \'$a\'`;
  } else {
    my $l_str = "&$queue(executable=\"$a\")(jobname=\"$J\")$inputFiles$outputFiles$env(arguments=''$g'')(stdout=\"$J.out\")(stderr=\"$J.err\")(cpuTime=\"$t\")(count=\"$n\")";
    $res_str = `$config{GRID_PREFIX}ngsub $cluster -e \'$l_str\'`;
#$r = `nohup $config{GRID_PREFIX}/ngsub $cluster -e \'$l_str\' 2>&1 /dev/null &`;
#$res .= "job rsl: ". $l_str. "\n";
  }
  if (my ($p) = $res_str =~ m/(gsiftp.+)/) {
    $res{id} = $p;
    ($res{H}) = $p =~ m/\/\/(.+):/;
  } else {
    $res{error} = $res_str;
  }
  if (defined $res{id}) {
    # Connect to mysql
    my $dsn = "DBI:mysql:database=$md;host=$mh;port=3306";
    my $dbh = DBI->connect($dsn, $ml, $mw,{'RaiseError' => 1});

    my $q = "INSERT grid_tasks (task_id, hostname, task_name, work_dir, login, middleware) values (\"$res{id}\", \"$res{H}\", \"$J\", \"$d\", \"$ENV{USER}\", \"arc\");";
    $dbh->do($q);
    $dbh->disconnect();
  }
  return \%res;
}

# ngstat -l -q
sub grid_resources {
  my @result;
#  my @res = `cat ~/grid-resources.txt`;
  my @res = `$config{GRID_PREFIX}ngstat -l -q`;

  my @allow_res = `$config{GRID_PREFIX}ngtest -R`;
#  my @allow_res = `cat ~/grid-allow.txt`;

# Parse allowed clusters
  my %allowed;
  my $cl_area = 0;
  foreach my $l (@allow_res) {
    if ($l =~ m/You are authorized at the following clusters/) {
      $cl_area = 1;
      next;
    }
    if ($cl_area && $l =~ m/^[ \t]*$/) {
      $cl_area = 0;
      next;
    }
    if ($cl_area) {
      for my $i (split(" ",$l)) {
        $allowed{$i} = 1;
      }
      next;
    }
  }

  my %resources;
  my $cluster;
  my %queue;
  my $desc_area = 0;
  my $queue_area = 0;
		
  # Parse cluster list
  foreach my $l (@res) {
    my $p;
    my $q;
    if ( ($l =~ m/^Cluster [\w.-]+/) || ($l =~ m/^Queue [\w.-]+/) || $l =~ m/^$/) {
    # Push queue to results
      if ($queue_area == 1) {
# print "Commit $queue{P}, l: $l, queue_area:$queue_area\n";
# print "$resources{'Number of CPUs'}\n" if defined $resources{'Number of CPUs'};
# print "$resources{'Number of Used CPUs'}\n" if defined $resources{'Number of Used CPUs'};

        my $y; my $w; my $d; my $h; my $m;
        my $t_str;
					
        if (defined $queue{'Max Wall Time'}) {
          $t_str = $queue{'Max Wall Time'};
        } elsif (defined $queue{'Max CPU Time'}) {
          $t_str = $queue{'Max CPU Time'};
	}
        if (defined $t_str) {
          ($y) = $t_str =~ m/(\d+) year/;
          ($w) = $t_str =~ m/(\d+) week/;
          ($d) = $t_str =~ m/(\d+) day/;
          ($h) = $t_str =~ m/(\d+) hour/;
          ($m) = $t_str =~ m/(\d+) minute/;
        }
        $y = 0 if not defined $y;
        $w = 0 if not defined $w;
        $d = 0 if not defined $d;
        $h = 0 if not defined $h;
        $m = 0 if not defined $m;
        $queue{'l'} = (((($y * 365) + $w*7 + $d)*24 + $h)*60+$m)*60;
        delete $queue{'Max Wall Time'};
        if ( $queue{'Status'} eq 'active' ) {
          if ($allowed{$queue{'H'}}) {
            $queue{'a'} = 'up';
          } else {
            $queue{'a'} = 'down';
          }
        } else {
          $queue{'a'} = 'fail';
        }
        delete $queue{'Status'};
					
        # Calculate Nodes
        $queue{'D'} = 0;
        if (defined $resources{'Number of Computers'}) {
          for my $i (split(/\n/, $resources{'Number of Computers'})) {
            my ($n) = $i =~ m/(\d+)$/;
            $queue{'D'} += $n;
          }
        }
        $resources{'Number of CPUs'} = 0 if not defined $resources{'Number of CPUs'};
        $resources{'Number of Used CPUs'} = 0 if not defined $resources{'Number of Used CPUs'};

        # Make idle queue
        $queue{'t'} = 'idle';
        $queue{'C'} = max($resources{'Number of CPUs'} - $resources{'Number of Used CPUs'}, 0);
        my %q_i = %queue;
        # Make alloc queue
        $queue{'t'} = 'alloc';
        $queue{'C'} = $resources{'Number of Used CPUs'};
        my %q_a = %queue;
        push @result, \%q_i;
        push @result, \%q_a;
        %queue = ();
        $queue_area = 0;
      }

    if (($p) = $l =~ m/^Cluster ([\w.-]+)/ ){
      $cluster = $p;
      $resources{'H'} = $p;
      %resources = ();
      $queue_area = 0;
#   print "Found cluster $cluster\n";
    }

    # Queue properties
    if (($p) = $l =~ m/^Queue ([\w.-]+)/ ){
      $queue_area = 1;
      # Load cluster properties to queue
      $queue{'H'} = $cluster;
      ($queue{'L'}) = $resources{'Location'} if defined $resources{'Location'};
      ($queue{'m'}) = $resources{'Memory on Each Node'} =~ m/^(\d+)/ if defined $resources{'Memory on Each Node'};
      $queue{'P'} = $p;
#	print "Found queue  $p"."\@"."$cluster\n";
#	printjson(\%queue);
#	print "\n";
      # Clean up
      $desc_area = 0;

      next;
    }

    next if not defined $cluster;

    if ( !$queue_area && (($p,$q) = $l =~ m/^  (\w[\w ]+): (.+)/)){
      $resources{$p} = $q;
      if ($desc_area) {
        $desc_area = 0;
      }				}
#  print "Found cluster property $p with $q\n";
#  printjson(\%resources);
#  print "\n";
      next;
    }
    if (!$queue_area && (($p) = $l =~ m/^ +([\w ]+):$/)) {
#print "Start multiline $p\n";
      $desc_area = $p;
      next;
    }
    if ($desc_area && (($p) = $l =~ m/^ +(.+)/)) {
#print "Continue multiline $desc_area with $p\n";
      $resources{$desc_area} = "" if not defined $resources{$desc_area};
      $resources{$desc_area} .= $p."\n";
      next;
    }

    if (!$desc_area && $queue_area && (($p,$q) = $l =~ m/ +(.*): +(.+)/)) {
      $queue{$p} = $q;
#	print "Found queue property\n";
#	printjson(\%queue);
#	print "\n";
      next;
    }
  }
  return \@result;
}


sub grid_queue {
  # parameter detail - detailed view
  my ($detail) = @_;
  $detail = 0 if not defined $detail;
  my $params;
  if ($detail == 0) {
    $params = "-l -s PREPARED -s ACCEPTED -s INLRMS:Q -s INLRMS:R -s SUBMITTING -s CANCELING -s INLRMS:E -s FINISHING";
  } else {
    $params = "-l -s PREPARED -s ACCEPTED -s INLRMS:Q -s INLRMS:R -s SUBMITTING -s CANCELING -s INLRMS:E -s FINISHING -s FINISHED -s FAILED";
  }
  my @result;
  my @res = `$config{GRID_PREFIX}ngstat $params`;
  my %job;
  foreach my $l (@res) {
    if (my ($p) = $l =~ m/Job information not found: (.+)/) {
      next;
    }
    if ( my ($p) = $l =~ m/^Job (.+)/) {
      if (defined $job{'i'}) {
        # Push job to results
        my %j_t = %job;
        push @result, \%j_t;
      }
      %job = ();
      $job{'i'} = $p;
#print "Found job $job{'id'}\n";
      next;
    }
    if (defined $job{'i'} && (my ($p,$q) = $l =~ m/ +([^:]+): (.*)/)) {
#print "P: $p, $q\n";
      if ($p eq 'Job Name') {
        $job{'j'} = $q;
      } elsif ($p eq 'Status') {
        $job{'t'} = 'CD'  if ($q eq 'FINISHED');
        $job{'t'} = 'F'   if ($q eq 'FAILED');
        $job{'t'} = 'PD'  if ($q eq 'INLRMS:Q' || $q eq 'PREPARED' || $q eq 'ACCEPTED' );
        $job{'t'} = 'R'   if ($q eq 'INLRMS:R');
        $job{'t'} = 'CG'  if ($q eq 'CANCELING' || $q eq 'FINISHING' || $q eq 'INLRMS:E');
        $job{'t'} = 'CA'   if ($q eq 'KILLED');
        $job{'t'} = 'TO'   if ($q eq 'DELETED');
      } elsif ($p eq 'Error') { 
        $job{'r'} = $q;
      } elsif ($p eq 'Cluster') { 
        $job{'H'} = $q;
      } elsif ($p eq 'Owner') { 
        ($job{'u'}) = $q =~m/CN=(.+)/;
      } elsif ($p eq 'Queue') { 
        $job{'P'} = $q;
      } elsif ($p eq 'Requested Number of CPUs') { 
        $job{'C'} = $q;
      } elsif ($p eq 'Submitted') { 
        $job{'b'} = str2time($q);
      } elsif ($p eq 'Completed') { 
        $job{'e'} = str2time($q);
      } else {
        $job{$p} = $q;
      }
      next;	
    }
  }
  if (defined $job{'i'}) {
    # Push last job to results
    my %j_t = %job;
    push @result, \%j_t;
  }
  return \@result;
}


sub grid_ls {
  my ($dir) = @_;
  my @list;
  my @res = `$config{GRID_PREFIX}ngls -l $dir`;
  foreach my $l (@res) {
    my %f; my $d; my $date;
    if (($f{'f'},$d,$f{'s'},$date) = $l =~ m/([^ ]+) ([^ ]+) (\d+) "(.+)"/ ) {
      $f{'s'} = int $f{'s'};
      $f{'t'} = str2time($date);
      $f{'d'} = 1;
      $f{'d'} = 0 if ( $d eq "file");
      push @list, \%f if ($f{'f'} ne "");
    }
  }
  return \@list;
}


sub grid_less {
  my($f,$o,$s) = @_;
  use File::Temp qw/ :POSIX /;
  my $tf = tmpnam();
  my $res = `$config{GRID_PREFIX}ngcp $f $tf 2>&1`;
  my $r = less($tf,$o,$s);

  system("rm -f $tf");
  return $r;
}


sub grid_put {
  my($f) = @_;
  use File::Temp qw/ :POSIX /;

  my $tf = tmpnam();
  open (FILE, ">", $tf);
  while (<STDIN>) {
    print FILE $_;
  }
  my $res = `$config{GRID_PREFIX}ngcp $tf $f 2>&1`;
  system ("rm -f $tf");
  print $res;
}


sub grid_get {
  my($d,$jid) = @_;
  my $res = `$config{GRID_PREFIX}ngget -dir $d $jid`;
}


sub grid_cp {
  my($sf_ref,$ddir) = @_;
  my $res = "";
  foreach my $f (@$sf_ref) {
    my ($fn) = $f =~ m/([^\/]+)$/;
    $res .= `$config{GRID_PREFIX}ngcp $f $ddir/$fn 2>&1`
  }
  return $res;
}

sub grid_cancel {
  my($jid) = @_;
  my $res = `$config{GRID_PREFIX}ngkill $jid`;
  return $res;
}


# Definitions

#our $grid_get = "$config{GRID_PREFIX}/ngget";

sub grid_get_str {
  my($d,$jid) = @_;
  return "$config{GRID_PREFIX}ngget -dir $d $jid";
}

sub grid_clean_str {
  my($jid) = @_;
  return "$config{GRID_PREFIX}ngclean -f $jid";
}

1;
