#
#   Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

use File::Temp qw/ :POSIX /;

sub user_keys {
  my ($s,$l,$h) = @_;
  
  !$s && return "server not specified";
  !$l && return "home base not specified";
  !$h && return "server not specified";
  
  my $admprefix="ssh root@".$s;

  my $res = "Making home dir for $l\n";
  my $id  = int `$admprefix id -u $l`;
  my $gid = int `$admprefix id -g $l`;

  my $cmd_str = 'if test -d /etc/skel -a \! -d '. "$h/$l". '\; then cp -r /etc/skel '. "$h/$l". ' \; else mkdir -p '. "$h/$l". '\; fi';
  $res .= `$admprefix $cmd_str`;

  $res .= `$admprefix chown -R $id\:$gid $h/$l`;
  $cmd_str = "su - $l -c ". '\"mkdir -p .ssh\"';
  $res .= `$admprefix $cmd_str`;

  $cmd_str = "su - $l -c ". '\"ssh-keygen -t rsa -N \\\'\\\' -f .ssh/id_rsa\"';
  $res .= `$admprefix $cmd_str`;

  $cmd_str = "su - $l -c ". '\"cat .ssh/id_rsa.pub \> .ssh/authorized_keys \&\& chmod 700 .ssh \&\& chmod 644 .ssh/authorized_keys\"';
  $res .= `$admprefix $cmd_str`;

  return $res;
}


sub ssh_copy_id {
  my ($s, $l, $w) = @_;
  # s - remote server
  # l - remote login
  # w - user password

  my $pub_key = `cat ~/.ssh/id_*.pub`; 
  
  my ($fh, $w_script) = tempfile( "askpw_XXXXXX", SUFFIX => '.sh');
  print $fh "#!/bin/bash\n";
  print $fh "echo $w\n";
  close($fh);
  system("chmod u+rx $w_script");

  # Env for ssh 
  $ENV{SSH_ASKPASS}=$w_script;

  # Set dummy display 
  $ENV{DISPLAY} = "dummydisplay:0";
  
  my $res = `setsid ssh -o StrictHostKeyChecking=no -o ConnectTimeout=2 $s \'umask 077; test -d ~/.ssh || mkdir ~/.ssh ; echo \"$pub_key\" >> ~/.ssh/authorized_keys\'`;
  unlink($w_script);
  return $res;
}

1;
