/*#   Copyright 2012 http://scms.pro/
#     Copyright 2012 Andrii Golovynskyi
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
*/

/*
 * SCMS Hydra Library
 */


#include <netdb.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#define CLUSTERCTL	"127.0.0.1"

static int ctlport = 9010;
static int cl;
static int uid;

static const char *hostname = CLUSTERCTL, *user, *task;

#include "utils.h"
#include "privatekey.h"

static void help( void) {

    fprintf( stdout, "hctl -v -i ipaddr -u username command\n");
    exit(0);
}

#define ERROR(args...) do{ fprintf( stderr, args); }while(0);


void jsonerror(char * s) {
  fprintf( stderr, "{\"Hctl Error\":\"%s\"}", s);
}


static void openserver( void) {

    struct sockaddr_in addr;
    struct hostent *hent;
    
    if( !(hent = gethostbyname( hostname))) {
	    ERROR( "hctl: cannot get hostaddr of hydrad server\n");
	    exit( 11);
    }

    memset( &addr, 0, sizeof( addr));
    addr.sin_family = AF_INET;
    addr.sin_addr = *(struct in_addr *)*hent->h_addr_list;
    addr.sin_port = htons( ctlport);
    
    if( (cl = socket( AF_INET, SOCK_STREAM, 0)) < 0) {
	    ERROR( "hctl: error_socket()\n");
	    exit( 12);
    }
    if( connect( cl, (struct sockaddr*)&addr, sizeof( addr)) < 0) {
	    ERROR( "hctl: error_connect() to hydrad server\n");
	    exit( 13);
    }
}


int main( int ac, char **av) {

    char *s, str[ 4194304];
    fd_set fdset;
    struct timeval tv;
    struct sockaddr_in addr;
    struct passwd *pswd;
    int c, len;
    
    int reverse = 0;

    uid = geteuid();

    ac--; av++;
    int ufl=0;
    int ifl=0;
    
    while( ac) {
	s = *av;
	if( *s != '-') break;
        av++; ac--;
        c = s[1];
	switch( c) {
	    case 'v':
		verbose++;
		continue;
	    case 'u':
		user = *av++; ac--;
		ufl=1;
		continue;
	    case 'i':
		hostname = *av++; ac--;
		ifl=1;
		continue;
	    default:
		help();
	}
    }

    if(!ufl || !ifl || !ac) 
	help();
    
    
    openserver();
     
    int type;

    if( (read_str( cl, str, sizeof( str), &type) != 4) || strcmp( str, "Yes?")) {
        ERROR("hctl: Hydrad server does not give \"Yes?\" promt: \"%s\"\n",str);
        return 1;
    }
    if( write_str( cl, "hydra", 5, TYPE_CTL) != 5) {
        ERROR("hctl: Failed to write \"hydra\" identifier\n");
        return 1;
    }
//TODO ADD HANDSHAKE HERE
    if( (len = read_str( cl, str, sizeof( str), &type)) <= 0) {
        ERROR("hctl: Handshake1 returned 0 bytes or error\n");
        return 1;
    }
// Cheking against HANDSHAKE1
    if( strcmp( str, HANDSHAKE1)) {
        ERROR("hctl: Handshake1 failed\n");
        return 1;
    }
    if( write_str( cl, HANDSHAKE2, HANDSHAKE2LEN, TYPE_CTL) != len) {
        ERROR("hctl: Sending Handshake2 to hydrad failed");
        return 1;
    }

    if( !read_ok( cl)) {
        ERROR("hctl: Hydrad reported that Handshake2 failed");
        return 1;
    }

    len = snprintf( str, sizeof( str), "user %s", user);
    if( (write_str( cl, str, len, TYPE_CTL) <= 0) || !read_ok( cl)) {
        ERROR("hctl: Sending of user failed, possibly hydrad does not accepted user %s", user);
        return 1;
    }
  // Sending command and it's parameters
  if( ac) {
	  len = 0;
    while( ac) {
	    s = *av++;
	    ac--;
	    len += snprintf( str + len, sizeof( str) - len, ac ? "%s " : "%s", s);
    }
    s = str;
  } else {
    help();
  }
  write_str( cl, s, len, TYPE_CTL);
  if( !read_ok( cl)) {
    ERROR("hctl: Failed to send command and it's parameters to hydrad");
    return 1;
  }
  // Write data to pipe first  
  const int readchunk = 4000000;
  int l;

  if ( !isatty(0)) {

    if( !freopen( NULL, "rb", stdin)) {
      perror( "freopen");
      ERROR("hctl: Failed to open stdin stream");
      exit(1);
    }

    while( !feof( stdin)) {
      l = fread( str, sizeof(char), readchunk, stdin);
      if( l <= 0)
        break;
      usleep(2);
      if (write_str( cl, str, l, TYPE_IN) <= 0)
        break;
    }
  }
  write_str( cl, ".DONE.", 0, TYPE_CTL);

  type = 0;
  // Then read results of command execution
  if( verbose) fprintf( stderr, "Reading results of command execution\n");
  while( l = read_str( cl, str, sizeof( str), &type)) {
    if( l <= 0) {
      ERROR( "Get %d result from read_str\n");
      break;
    }
    int res;
    switch(type){
    case TYPE_OUT:
      res = fwrite( str, sizeof(char), l, stdout);
      break;
    case TYPE_ERR:
      res = fwrite( str, sizeof(char), l, stderr);
      break;
    default:
      ERROR( "Get wrong packet type %d\n", type);
    }
    if( res <= 0) {
      ERROR( "Error writing packet to output\n");
      break;
    }
  }

 return 0;
}

